/***************************************************************************
 $RCSfile: api.cpp,v $
                             -------------------
    cvs         : $Id: api.cpp,v 1.22 2003/03/04 17:34:57 derzach Exp $
    begin       : Sat Jun 08 2002
    copyright   : (C) 2001 by Martin Preuss
    email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "api.h"
#include <hbcistring.h>
#include <accountimpl.h>
#include <mediumddv.h>
#include <mediumkeyfile.h>
#ifdef WITH_RSACARD
# include <mediumrsacard.h>
#endif
#include <openhbci/connection.h>



namespace HBCI {


API::API(bool readonly,
		 bool retrievalonly)
    :Hbci(readonly, retrievalonly)
{
    _loader=new Loader(this);
    _loader.setDescription("API::_loader");
    _loader.setObjectDescription("Loader");
    _monitor=new ProgressMonitor();
    _monitor.setDescription("API::_monitor");
    _monitor.setObjectDescription("ProgressMonitor");
    _queue=new Outbox(this);
    _queue.setDescription("API::_queue");
    _queue.setObjectDescription("Outbox");
}


API::~API() {
}


Pointer<ProgressMonitor> API::monitor() const {
    return _monitor;
}


void API::setMonitor(Pointer<ProgressMonitor> monitor){
    _monitor=monitor;
}


void API::addBank(Pointer<Bank> b){
    for(list<Pointer<Bank> >::const_iterator iter=_banks.begin();
        iter!=_banks.end(); iter++) {
        /* first check if the very same object already exists.
         * It's not a problem if it does, since it is the SAME.
         */
        if (*iter==b) {
#if DEBUGMODE>0
            fprintf(stderr,
                    "Will not add bank, since it already IS listed.\n");
#endif
            return;
        }
        /* then check if there already is an object with the same id data.
         * This WOULD be a problem, since that would corrupt the list.
         * The second occurence of an equal account could rarely be accessed.
         */
        if ((*iter).ref().countryCode()==b.ref().countryCode())
            if ((*iter).ref().bankCode()==b.ref().bankCode())
                throw Error("BankImpl::addBank()",
                                ERROR_LEVEL_NORMAL,
                                0,
                                ERROR_ADVISE_DONTKNOW,
                                "bank already exists.");
    } // for

    // otherwise simply add the bank
    _banks.push_back(b);
}


Pointer<Bank> API::findBank(int country,
                                        const string &instcode) const {
    for(list<Pointer<Bank> >::const_iterator iter=_banks.begin();
        iter!=_banks.end(); iter++) {
        if ((*iter).ref().countryCode()==country)
            if ((*iter).ref().bankCode()==instcode)
                return (*iter);
    }
    return 0;
}


Pointer<Account> API::findAccount(int country,
                                              const string &instcode,
                                              const string &accnr,
                                              const string &suffix) const {
  // ok, this is crappy:
  // there might be accounts that have another institute id than the 
  // bank they belong to (Hypovereinsbank) !?!
  // So we need to loop over all banks, check if any of their accounts
  // match "accnr"
  for(list<Pointer<Bank> >::const_iterator iter=_banks.begin();
	  iter!=_banks.end(); iter++) {

	Pointer<Account> account = 
	  (*iter).ref().findAccount(accnr, suffix);
	
	if (account.isValid() &&
		account.cast<AccountImpl>().ref().instituteCode() == instcode &&
		account.cast<AccountImpl>().ref().countryCode() == country) {
	  return account;
	}
  }

  // not found
  return NULL;
}


Pointer<User> API::findUser(int country,
                                        const string &instcode,
                                        const string &user) const {
    Pointer<Bank> bank;

    bank=findBank(country, instcode);
    if (!bank.isValid())
        return 0;
    return bank.ref().findUser(user);
}


Pointer<Customer> API::findCustomer(int country,
                                                const string &instcode,
                                                const string &cust) const {
    Pointer<Bank> bank;

    bank=findBank(country, instcode);
    if (!bank.isValid())
        return 0;
    return bank.ref().findCustomer(cust);
}

int API::totalAccounts () const 
{
    list<Pointer<Bank> >::const_iterator iter, end;
    int ret = 0;

    end = bankList().end();
    for ( iter = bankList().begin(); iter != end; iter++ ) {
	ret += (*iter).ref().accounts().size();
    }
    return ret;
}
int API::totalUsers () const 
{
    list<Pointer<Bank> >::const_iterator iter, end;
    int ret = 0;

    end = bankList().end();
    for ( iter = bankList().begin(); iter != end; iter++ ) {
	ret += (*iter).ref().users().size();
    }
    return ret;
}


Error API::loadEnvironment(const string &f,
                                   unsigned int fl){
    SimpleConfig cfg;
    Error err;

    cfg.setMode(HBCIAPP_CONFIG_MODE);
    err=cfg.readFile(f);
    if (!err.isOk())
        return err;
    err=_loader.ref().loadAll(cfg,cfg.root(),fl);

    return err;
}


Error API::saveEnvironment(const string &f,
                                   unsigned int fl) const {
    SimpleConfig cfg;
    Error err;

    cfg.setMode(HBCIAPP_CONFIG_MODE);
    err=_loader.ref().saveAll(cfg,cfg.root(),fl);
    if (!err.isOk())
        return err;
    err=cfg.writeFile(f);
    if (!err.isOk())
        return err;

    return err;
}


Error API::postProcessInitJob(JOBDialogInit &job) {
    Pointer<userParams> upd;
    Pointer<bankParams> bpd;
    Pointer<Bank> bank;
    Pointer<Customer> cust;
    Pointer<RSAKey> signKey;
    Pointer<RSAKey> cryptKey;

    cust=job.owner();
    User &user = cust.ref().user().ref();

    bpd = job.bpd();
    upd = job.upd();
    signKey = job.serverSignKey();
    cryptKey = job.serverCryptKey();

    // check for BPD
    if (bpd.isValid()) {
        // check if the bank already exists
        bank=findBank(bpd.ref().countryCode(),
                      bpd.ref().instituteCode());
        if (bank.isValid()) {
            // it does, so set the new values
            // do not overwrite server (WORKAROUND)
            bpd.ref().setAddr(bank.ref().addr());
            dynamic_cast<BankImpl&>(bank.ref())=bpd.ref();
        }
        else {
            // does not exists, add it
            addBank(new BankImpl(this,bpd.ref()));
        }
    }
    // check for keys
    if (cryptKey.isValid())
        cust.ref().user().ref().medium().cast<MediumRDHBase>().ref()
            .setInstituteCryptKey(cryptKey);
    if (signKey.isValid())
        cust.ref().user().ref().medium().cast<MediumRDHBase>().ref()
            .setInstituteSignKey(signKey);
    job.clearKeys();

    // check for UPD
    if (upd.isValid()) {
        list<accountParams> accs;
        list<accountParams>::const_iterator ait;
        Pointer<Account> acc;
        Pointer<AccountImpl> aimpl;

        bank=user.bank();
        user.setVersion(upd.ref().version());
		user.setKnowsSupportedJobs(upd.ref().knowsSupportedJobs());

        accs=upd.ref().accounts();
        for (ait=accs.begin(); ait!=accs.end(); ait++) {
            acc=bank.ref().findAccount((*ait).accountNumber(),
                                       (*ait).accountSuffix());
            if (acc.isValid()) {
                // account exists, so set new basic params
		// Downcast of existing account
		aimpl = acc.cast<AccountImpl>();
		aimpl.ref() = *ait;
                //dynamic_cast<AccountImpl&>(acc.ref())=*ait;
                aimpl.ref().addAuthorizedCustomer(cust);
            }
            else {
                // otherwise simply add it
                Pointer<Account> acc;
                aimpl=new AccountImpl(bank,*ait);
                aimpl.ref().addAuthorizedCustomer(cust);
		// Upcast to store only the reference to the base
		// class.
		acc = aimpl.cast<Account>();
                acc.setObjectDescription("Account");
                bank.ref().addAccount(acc);
            }
        } // for
    }

    return Error();
}


Error
API::processInstituteMessages(Pointer<MessageQueue> mbox){
    list<instituteMessage>::const_iterator it;
    Pointer<Bank> bank;
    bool iserr=false;

    // check all messages and copy them to the appropriate bank
    for (it=mbox.ref().bankMessages().begin();
         it!=mbox.ref().bankMessages().end();
         it++) {
        bank=findBank((*it).country(),(*it).bankCode());
        if (!bank.isValid())
            iserr=true;
        else {
            // store message in the bank
            BankImpl &ba = dynamic_cast<BankImpl&> (bank.ref());
            ba.addInstituteMessage(*it);
        }
    } // for
    if (iserr)
        return Error("API::processInstituteMessages()",
                         ERROR_LEVEL_NORMAL,
                         0,
                         ERROR_ADVISE_DONTKNOW,
                         "bank does not exist.");
    else
        return Error();
}


Error API::_handleMessageQueue(Pointer<Connection> conn,
                                       Pointer<MessageQueue> mbox){
    Error err;

    try {
        if (mbox.ref().empty()) {
            // empty queue, finished ;-)
            _monitor.ref().actionStarted(ACT_SENDINGMESSAGE,
					 "Sending message (nothing to do)");
            _monitor.ref().actionFinished();
            _monitor.ref().actionStarted(ACT_WAITRESPONSE,
					 "Waiting for response (nothing to do)");
            _monitor.ref().actionFinished();
            return Error();
        }

        _monitor.ref().actionStarted(ACT_SENDINGMESSAGE,
				     "Sending message");
        // send this message
        if (!conn.ref().sendMessage(mbox))
	    return Error("API::_handleMessageQueue()",
			 ERROR_LEVEL_INTERNAL,
			 0,
			 ERROR_ADVISE_ABORT,
			 "could not send");
	_monitor.ref().actionFinished();
        _monitor.ref().actionStarted(ACT_WAITRESPONSE,
				     "Waiting for response");
        // await response
        if (!conn.ref().getResponse())
            return Error("API::_handleMessageQueue()",
                         ERROR_LEVEL_INTERNAL,
                         0,
                         ERROR_ADVISE_ABORT,
                         "nothing received");
        _monitor.ref().actionFinished();
    }
    catch(Error xerr) {
        err=xerr;
    }
    if (!err.isOk())
        _logMessage(1,"RESULT (handleMessage): "+err.errorString());
    //fprintf(stderr,"RESULT (handleMessage): %s\n",
    //        err.errorString().c_str());
    return err;
}


void API::_logMessage(int loglevel, const string &msg) {
    if (_monitor.isValid()) {
        if (loglevel<debugLevel())
            _monitor.ref().logMessage(msg);
    }
}


Error API::_handleJobQueue(const list<Pointer<OutboxJob> > &jq,
                                   Pointer<Connection> conn,
                                   Pointer<MessageQueue> mbox,
                                   bool dlg,
                                   bool chg){
    list<Pointer<OutboxJob> >::const_iterator it;
    Error err;
    int jobErrors=0;

    // handle the jobs
    for (it=jq.begin(); it!=jq.end(); it++) {
#if DEBUGMODE>0
        fprintf(stderr,"HANDLE_QUEUE: job is %s\n",
                (*it).ref().description().c_str());
#endif
        if ((*it).ref().isDialogJob()==dlg &&
            (*it).ref().status()==HBCI_JOB_STATUS_TODO) {
            _monitor.ref().jobStarted((*it).ref().type(),
				      (*it).ref().description(),
				      (*it).ref().messages()*3+(chg?2:1));

            // let the job create its messages and send them
            int i=-1;
            while ((*it).ref().stillMessagesToSend(++i)) {
                // create real jobs and add them to the message queue
                _monitor.ref().actionStarted(ACT_CREATEHBCIJOB,
					     "Creating hbci jobs "+
                                             String::num2string(i));
                if (!(*it).ref().createHBCIJobs(mbox,i)) {
                    // error on creation, break
                    jobErrors++;
                    break;
                }
                _monitor.ref().actionFinished();

                // now send the message queue to the server
                err=_handleMessageQueue(conn,mbox);
                if (!err.isOk())
                    return err;

                // during a dialog-job, new data may be returned (upd, bpd,
                // signaturekeys, system-id...)
                // some of them are necessary in the dialog-end-message
                // to finish the dialog correctly
                // if we are allowed to we let the job change this data
                // within the current context
                if (chg)
                    (*it).ref().commit(i);

                // reset mbox for next use
                mbox.ref().reset();
            } // while

            // all jobs handled ?
            if ((*it).ref().stillMessagesToSend(i))
                // no, stop handling this queue
                break;

            // mark job as done
            (*it).ref().setStatus(HBCI_JOB_STATUS_DONE);

            // evaluate this job
            _monitor.ref().actionStarted(ACT_CHKRESULT,
					 "Checking job result.");
            if (!(*it).ref().evaluate())
                jobErrors++;
            _monitor.ref().actionFinished();

            // commit changes if wanted
            if (chg) {
                _monitor.ref().actionStarted(ACT_UPDATESYSTEM,
					     "Updating system.");
                if (!(*it).ref().commit(-1))
                    jobErrors++;
                _monitor.ref().actionFinished();
            }
            _monitor.ref().jobFinished();
        }
    } // for

    if (jobErrors>0)
        return ( (jobErrors == 1) ?
		 err :
		 Error("API::_handleJobQueue()",
		       ERROR_LEVEL_NORMAL,
		       1,
		       ERROR_ADVISE_DONTKNOW,
		       "more than one job failed"));
    return Error();
}


Error API::_handleUserQueue(Pointer<customerQueue> q,
                                    Pointer<Connection> conn,
                                    bool chg){
    list<Pointer<OutboxJob> > jq;
    Error err;
    int jobErrors=0;
    int j;
    Pointer<Customer> cust;
    Pointer<MessageQueue> mbox;
    JOBDialogInit *initJobImpl;
    Pointer<Job> initJob;
    Pointer<Job> exitJob;

    // get queue size
    j=q.ref().sizeByStatus(HBCI_JOB_STATUS_TODO);
    if (!j)
        // nothing to be done
        return Error();

    // get corresponding customer
    cust=q.ref().customer();
    if (!cust.isValid())
        return Error("API::_handleUserQueue()",
                         ERROR_LEVEL_INTERNAL,
                         0,
                         ERROR_ADVISE_ABORT,
                         "no customer in customerQueue");

    if (Hbci::debugLevel()>2)
	fprintf(stderr,"HANDLEUSERQUEUE: customer is %s\n",
		cust.ref().custId().c_str());

    jq=q.ref().jobs();
    mbox=new MessageQueue(cust);

    // first handle all dialog jobs
    err=_handleJobQueue(jq,conn,mbox,true,chg);
    if (!err.isOk()) {
	jobErrors++;
	if (err.advise()==ERROR_ADVISE_ABORT)
            return err;
    }

    // get queue size
    j=q.ref().sizeByStatus(HBCI_JOB_STATUS_TODO);
    if (!j)
        // nothing to be done
        return Error();

    // otherwise handle non-dialog jobs
    _monitor.ref().jobStarted(JOB_OPENINGDIALOG,
			      "Opening dialog",chg ? 3:2);

    // create non anonymous dialog init job
    initJobImpl=new JOBDialogInit(cust,false);
    initJob = initJobImpl;
    mbox.ref().addJob(initJob);

    err=_handleMessageQueue(conn,mbox);
    if (!err.isOk())
        return err;

    /* commit possibly pending changes inflicted by the servers answer
     * only if wanted */
    if (chg) {
        _monitor.ref().actionStarted(ACT_UPDATESYSTEM,
				     "Updating system.");
        err=postProcessInitJob(*initJobImpl);
        if (!err.isOk())
            return err;
        // handle institute messages
        err=processInstituteMessages(mbox);
        if (!err.isOk())
            return err;
        _monitor.ref().actionFinished();
    }
    _monitor.ref().jobFinished();

    // reset mbox for next use
    mbox.ref().reset();

    // now handle normal jobs
    err=_handleJobQueue(jq,conn,mbox,false,chg);
    if (!err.isOk()) {
	jobErrors++;
	if (err.advise()==ERROR_ADVISE_ABORT)
	    return err;
    }

    // create exit job
    exitJob=new JOBDialogEnd(cust,mbox.ref().dialogId());
    mbox.ref().addJob(exitJob);

    _monitor.ref().jobStarted(JOB_CLOSINGDIALOG,
			      "Closing dialog",2);
    // send this message
    err=_handleMessageQueue(conn,mbox);
    _monitor.ref().jobFinished();
    if (jobErrors>0)
        return ( (jobErrors == 1) ?
		 err :
		 Error("API::_handleUserQueue()",
		       ERROR_LEVEL_NORMAL,
		       1,
		       ERROR_ADVISE_DONTKNOW,
		       "more than one job failed"));
    return err;
}


Error API::_handleBankQueue(Pointer<bankQueue> q, bool chg){
    list<Pointer<customerQueue> > cq;
    list<Pointer<customerQueue> >::const_iterator it;
    Pointer<Connection> conn;
    int j;
    Pointer<Bank> bank;
    Error err;
    int jobErrors=0;
    int connectionErrors=0;

    // get queue size
    j=q.ref().sizeByStatus(HBCI_JOB_STATUS_TODO);
    if (!j)
        // nothing to be done
        return Error();

    // get corresponding bank
    bank=q.ref().bank();
    if (!bank.isValid())
        return Error("API::_handleBankQueue()",
                         ERROR_LEVEL_INTERNAL,
                         0,
                         ERROR_ADVISE_ABORT,
                         "no bank in bankQueue");
#if DEBUGMODE>0
    fprintf(stderr,"HANDLEBANKQUEUE: bank is %s\n",
	    bank.ref().bankCode().c_str());
#endif

    cq=q.ref().customerQueues();
    for (it=cq.begin(); it!=cq.end(); it++) {
        _monitor.ref().jobStarted(JOB_OPENINGNETWORK,
				  "Opening network connection",1);
        _monitor.ref().actionStarted(ACT_CONTACTINGSERVER,
				     "Contacting server "+
                                     bank.ref().addr());
        conn=new Connection(this,bank.ref().addr());
	err = conn.ref().open();
	if ( !err.isOk() ){
            _logMessage(1,"Could not connect to "+bank.ref().name()+
                        " ("+bank.ref().bankCode()+")");
            // skip all jobs in this queue
            connectionErrors++;
            break;
        } // if !open
        _monitor.ref().actionFinished();
        _monitor.ref().jobFinished();

        // handle user jobs
        err=_handleUserQueue(*it, conn, chg);
        if (!err.isOk()) {
            _logMessage(1,"HandleUserQueue: "+err.errorString());
            jobErrors++;
	    if (err.advise()==ERROR_ADVISE_ABORT)
		break;
	}
        // close physical connection
        _monitor.ref().jobStarted(JOB_CLOSINGNETWORK,
				  "Closing network connection",1);
        _monitor.ref().actionStarted(ACT_CLOSECONNECTION,
				     "Closing connection");
        if (!conn.ref().close())
            connectionErrors++;
        _monitor.ref().actionFinished();
        _monitor.ref().jobFinished();
    } // for

    // check for errors
    if (jobErrors>0)
        return ( (jobErrors == 1) ?
		 err :
		 Error("API::_handleBankQueue()",
		       ERROR_LEVEL_NORMAL,
		       1,
		       ERROR_ADVISE_DONTKNOW,
		       "more than one job failed"));
    if (connectionErrors > 0)
        return ( (connectionErrors == 1) ?
		 err :
		 Error("API::_handleBankQueue()",
		       ERROR_LEVEL_NORMAL,
		       2,
		       ERROR_ADVISE_DONTKNOW,
		       "some connections failed"));
    return Error();
}


Error API::_handleQueue(Pointer<Outbox> q, bool chg){
    list<Pointer<bankQueue> > bq;
    list<Pointer<bankQueue> >::const_iterator it;
    int j;
    int jobErrors=0;
    Error err;

    // get queue size
    j=q.ref().sizeByStatus(HBCI_JOB_STATUS_TODO);
    if (!j)
        // nothing to be done
        return Error();
    bq=q.ref().bankQueues();

    // add jobs for connection opening and closing
    j+=q.ref().bankCount()*2;

    // add jobs for dialog opening and closing
    if (!q.ref().allDialogJobs())
	j+=q.ref().customerCount()*2;

    _monitor.ref().transactionStarted(TRANS_JOBQUEUE,
				      "Executing job queue",j);
    for (it=bq.begin(); it!=bq.end(); it++) {
        err=_handleBankQueue(*it,chg);
        if (!err.isOk()){
            _logMessage(1,"Bank: "+err.errorString());
            jobErrors++;
        }
    }

    _monitor.ref().transactionFinished();
    
    if (jobErrors>0)
        return ( (jobErrors == 1) ?
		 err :
		 Error("API::_handleQueue()",
		       ERROR_LEVEL_NORMAL,
		       1,
		       ERROR_ADVISE_DONTKNOW,
		       "more than one job failed"));
    return Error();
}


Error API::addJob(Pointer<OutboxJob> job){
    _queue.ref().addJob(job);

    return Error();
}


Error API::executeQueue(bool chg){
    Error err;

    err= _handleQueue(_queue,chg);
    return err;
}


list<Pointer<Account> > 
API::getAccounts(int country,
		     const string &instid,
		     const string &accnr) const {
    list<Pointer<Bank> >::const_iterator it;
    list<Pointer<Account> > accs;
    list<Pointer<Account> >::const_iterator ait;;

    for (it=_banks.begin(); it!=_banks.end();it++) {
        if ((country==0 || (*it).ref().countryCode()) &&
            (-1!=parser::cmpPattern((*it).ref().bankCode(),
                                      instid,false))) {

            // add matching accounts to mine
            for (ait=(*it).ref().accounts().begin();
                 ait!=(*it).ref().accounts().end(); ait++)
                if (-1!=parser::cmpPattern((*ait).ref().accountId(),
                                             accnr,false))
                    accs.push_back(*ait);
        } // if bank matches
    } // for
    return accs;
}


list<Pointer<User> > 
API::getUsers(int country,
                  const string &instid,
                  const string &userid) const {
    list<Pointer<Bank> >::const_iterator it;
    list<Pointer<User> > custs;
    list<Pointer<User> >::const_iterator cit;;

    for (it=_banks.begin(); it!=_banks.end();it++) {
        if ((country==0 || (*it).ref().countryCode()) &&
            (-1!=parser::cmpPattern((*it).ref().bankCode(),
                                      instid,false))) {
            // add matching customers to mine
            for (cit=(*it).ref().users().begin();
                 cit!=(*it).ref().users().end(); cit++)
                if (-1!=parser::cmpPattern((*cit).ref().userId(),
                                             userid,false))
                    custs.push_back(*cit);
        } // if bank matches
    } // for
    return custs;
}


list<Pointer<Customer> >
API::getCustomers(int country,
                      const string &instid,
                      const string &custid) const {
    list<Pointer<Bank> >::const_iterator it;
    list<Pointer<User> >::const_iterator uit;;
    list<Pointer<Customer> >::const_iterator cit;;
    list<Pointer<Customer> > custs;

    // iterate through banks
    for (it=_banks.begin(); it!=_banks.end();it++) {
        if ((country==0 || (*it).ref().countryCode()) &&
            (-1!=parser::cmpPattern((*it).ref().bankCode(),
                                      instid,false))) {
            // iterate through users
            for (uit=(*it).ref().users().begin();
                 uit!=(*it).ref().users().end(); uit++) {
                // add matching customers to mine
                for (cit=(*uit).ref().customers().begin();
                     cit!=(*uit).ref().customers().end(); cit++)
                    if (-1!=parser::cmpPattern((*cit).ref().custId(),
                                                 custid,false))
                        custs.push_back(*cit);
            } // for user
        } // if bank matches
    } // for bank
    return custs;
}


Pointer<Medium> API::findMedium(const string &name) const {
    list<Pointer<Bank> >::const_iterator bit;
    list<Pointer<User> >::const_iterator cit;
    Pointer<Medium> medium;

    for (bit=_banks.begin(); bit!=_banks.end();bit++) {
        // add matching accounts to mine
        for (cit=(*bit).ref().users().begin();
             cit!=(*bit).ref().users().end(); cit++) {
            medium=(*cit).ref().medium();
            if (medium.isValid())
                if (medium.ref().mediumName()==name)
                    return medium;
        } // for

    } // for

    // not found
    return 0;
 
}


Pointer<Medium> API::createNewMedium(int country,
                                     const string &bankId,
                                     const string &userId,
                                     const string &name,
                                     MediumType mt) const {
    Pointer<Medium> medium;
    Error err;

    // first check arguments
    if (bankId.empty() || userId.empty())
        throw Error("API::createNewMedium()",
                    ERROR_LEVEL_NORMAL,
                    0,
                    ERROR_ADVISE_DONTKNOW,
                    "empty argument");

    // check security mode, create medium
    if (mt==MediumRDHFile) {
        // name empty ?
        if (name.empty())
            throw Error("API::createNewMedium()",
                        ERROR_LEVEL_NORMAL,
                        0,
                        ERROR_ADVISE_DONTKNOW,
                        "need name"
                        "for RDH mode the name contains the path for the keyfile");

        // check if file already exists
        File f(name);
        if (!f.accessFile(FILE_ACCESS_EXIST).isOk()) {
            // File does not exist

            // create medium; Create new MediumRDH object
            MediumKeyfile *mediumtmp = new MediumKeyfile(this,name);
            // Store it in the upcast version
            medium = mediumtmp;
            // For convenience, use a reference in the following
            MediumKeyfile &mediumrdh = *mediumtmp;

            string pin;
            if (!interactor().ref().msgInputPin (0,
                                                 pin,
                                                 mediumrdh.minPinSize,
                                                 true))
                throw Error("API::createNewMedium()",
                            ERROR_LEVEL_NORMAL,
                            0,
                            ERROR_ADVISE_DONTKNOW,
                            "could not get pin");
            // create context
            if (!mediumrdh.createMedium(country,
                                        bankId,
                                        userId,
                                        pin))
                throw Error("API::createNewMedium()",
                            ERROR_LEVEL_NORMAL,
                            0,
                            ERROR_ADVISE_DONTKNOW,
                            "could not create medium");

	    // mount medium without customer
	    err=mediumrdh.mountMedium(0, pin);
	    if (!err.isOk())
		throw Error("API::createNewMedium()",err);

            // select context
            if (!mediumrdh.selectContext(country, bankId, userId))
                throw Error("API::createNewMedium()",
                            ERROR_LEVEL_NORMAL,
                            0,
                            ERROR_ADVISE_DONTKNOW,
                            "could not select context");

            // create keys
            if (!mediumrdh.createUserKeys())
                throw Error("API::createNewMedium()",
                            ERROR_LEVEL_NORMAL,
                            0,
                            ERROR_ADVISE_DONTKNOW,
                            "could not create keys");
            // unmount medium
            mediumrdh.unmountMedium(pin);
        } // if not exists
        else
            // File exists, so: create medium
            medium=new MediumKeyfile(this,name);

        // finished

    } // RDH
    else if (mt==MediumDDVCard) {
#ifndef HAS_LIBCHIPCARD
        throw Error("API::createNewMedium()",
                    ERROR_LEVEL_NORMAL,
                    0,
                    ERROR_ADVISE_DONTKNOW,
                    "you need libchipcard for DDV mode",
                    "get it from www.libchipcard.de");
#else
        // the case is simple here, all data already exists on the card
        medium=new MediumDDV(this,
                             name);
#endif
    }
    else if (mt==MediumRDHCard) {
#ifndef HAS_LIBCHIPCARD
      throw Error("API::createNewMedium()",
		  ERROR_LEVEL_NORMAL,
		  0,
		  ERROR_ADVISE_DONTKNOW,
		  "you need libchipcard v0.7 or better for RSA cards",
		  "get it from www.libchipcard.de");
#else
# if (k_CHIPCARD_VERSION_MAJOR<1 && k_CHIPCARD_VERSION_MINOR<7)
      throw Error("API::createNewMedium()",
		  ERROR_LEVEL_NORMAL,
		  0,
		  ERROR_ADVISE_DONTKNOW,
		  "you need a newer version of libchipcard (v0.7 or better "
		  "for RSA cards",
		  "get it from www.libchipcard.de");
# else
      // the case is simple here, all data already exists on the card
#  ifdef WITH_RSACARD
      medium=new MediumRSACard(this,
			       name);
#  endif // WITH_RSACARD
# endif // correct libchipcard version
#endif // HAS_LIBCHIPCARD
    }
    else
        throw Error("API::createNewMedium()",
                    ERROR_LEVEL_NORMAL,
                    0,
                    ERROR_ADVISE_DONTKNOW,
                    "bad security mode");
    return medium;
}


Pointer<User> API::userFactory(Pointer<Bank> b,
                               Pointer<Medium> m,
                               const string &userid,
                               int version,						   
                               const string &userName,
							   bool knowsSupportedJobs) {
    return new User(b,m,userid,version, userName, knowsSupportedJobs);
}


Pointer<Customer> API::customerFactory(Pointer<User> u,
                                       const string &id,
                                       const string &role) {
    return new Customer(u,id,role);
}


Pointer<Bank> API::bankFactory(int country,
                               const string &bankCode,
                               const string &server,
                               int hbciVersion) const {
  Pointer<Bank> newBank = 
	new BankImpl(this,country,bankCode,server,hbciVersion);

  // set the default-language to german
  newBank.ref().setLanguage(HBCI_LANGUAGE_GERMAN);

  return newBank;
}


void API::removeBank(Pointer<Bank> b) {
    list<Pointer<Bank> >::iterator it1;

    // find bank in list and remove it
    for(it1=_banks.begin();
        it1!=_banks.end(); it1++) {
        if ((*it1)==b) {
            _banks.erase(it1);
            break;
        }
    } // for
}



Pointer<Account>
API::accountFactory(Pointer<Bank> b,
                    const string &accountId,
                    const string &accountSubId) {
    return new AccountImpl(b,accountId,accountSubId);
}


Pointer<Medium> API::mediumFactory(const string &name,
                                   MediumType t) const {
    Pointer<Medium> m;

    switch (t) {
    case MediumRDHFile:
        if (Hbci::debugLevel()>2)
            cerr<<"Creating RDHFile medium\n";
        m=new MediumKeyfile(this,name);
        break;

    case MediumDDVCard:
#ifdef HAS_LIBCHIPCARD
        if (Hbci::debugLevel()>2)
            cerr<<"Creating DDVCard medium\n";
        m=new MediumDDV(this,name);
#endif
        break;

    case MediumRDHCard:
#ifdef HAS_LIBCHIPCARD
# if (k_CHIPCARD_VERSION_MAJOR>0 || k_CHIPCARD_VERSION_MINOR>6)
#  ifdef WITH_RSACARD
        if (Hbci::debugLevel()>2)
            cerr<<"Creating RDHCard medium\n";
	m=new MediumRSACard(this,name);
#  endif // WITH_RSACARD
# endif // correct libchipcard version
#endif // HAS_LIBCHIPCARD
        break;

    default:
        if (Hbci::debugLevel()>2)
            cerr<<"Not creating any medium\n";
        break;
    } // switch

    // finally return the medium
    return m;
}


list<Pointer<OutboxJob> > API::queuedJobs() const {
    return _queue.ref().jobs();
}


void API::clearQueueByResult(OutboxJob_Result result) {
    _queue.ref().removeByResult(result);
}


void API::clearQueueByStatus(OutboxJob_Status status){
    _queue.ref().removeByState(status);
}


void API::removeQueuedJob(Pointer<OutboxJob> job){
    _queue.ref().removeJob(job);
}


} // namespace HBCI


// ------------------ C Wrappers

// Helpers for non-deleting Pointer
static HBCI::Pointer<HBCI_Bank> bankPointer(HBCI_Bank *b)
{
    HBCI::Pointer<HBCI_Bank> bp = b;
    bp.setAutoDelete(false);
    return bp;
}
static HBCI::Pointer<HBCI_User> userPointer(HBCI_User *u)
{
    HBCI::Pointer<HBCI_User> up = u;
    up.setAutoDelete(false);
    return up;
}

// class HBCI_API
HBCI_API *HBCI_API_new(int readonly, int retrievalOnly)
{
    return new HBCI_API(readonly, retrievalOnly);
}
void HBCI_API_delete(HBCI_API *h)
{
    delete h;
}
HBCI_Hbci *HBCI_API_Hbci(HBCI_API *h)
{
    return h;
}
void HBCI_API_setMonitor(HBCI_API *h, HBCI_ProgressMonitor *m, int autoDelete)
{
    assert(h);
    assert(m);
    HBCI::Pointer<HBCI::ProgressMonitor> mp = m;
    mp.setAutoDelete(autoDelete);
    h->setMonitor(mp);
}
const list_HBCI_Bank *HBCI_API_bankList(const HBCI_API *h)
{
    assert(h);
    return &(h->bankList());
}
HBCI_Bank *HBCI_API_findBank(const HBCI_API *h, int country, 
			     const char *bankCode)
{
    assert(h);
    return (h->findBank(country, string(bankCode ? bankCode : ""))).ptr();
}
/*HBCI_Account *HBCI_API_findAccount(const HBCI_API *h, int country,
				   const char *bankCode,
				   const char *accnr)
{
    assert(h);
    return (h->findAccount(country, string(bankCode), string(accnr))).ptr();
    }*/
/*HBCI_Account *HBCI_API_findAccountWSuf(const HBCI_API *h, int country,
				       const char *bankCode,
				       const char *accnr,
				       const char *suffix)
{
    assert(h);
    return (h->findAccount(country, string(bankCode), string(accnr),
			   string(suffix))).ptr();
			   }*/
/*HBCI_User *HBCI_API_findUser(const HBCI_API *h,
			     int country,
			     const char *bankCode,
			     const char *userId)
{
    assert(h);
    return (h->findUser(country, string(bankCode), string(userId))).ptr();
    }*/
HBCI_Medium *HBCI_API_findMedium(const HBCI_API *h,
				 const char *name)
{
    assert(h);
    return (h->findMedium(string(name ? name : ""))).ptr();
}
HBCI_Error *HBCI_API_addBank(HBCI_API *h, HBCI_Bank *b, int autoDelete)
{
    assert(h);
    assert(b);
    HBCI::Pointer<HBCI_Bank> bp = b;
    bp.setAutoDelete(autoDelete);
    try {
	h->addBank(bp);
	return 0;
    }
    catch(HBCI_Error err) {
	return new HBCI_Error(err);
    }
    catch (...) {
	fprintf(stderr, "HBCI_API_addBank: Caught unknown exception.\n");
	return new HBCI_Error("HBCI_API_addBank",
			      ERROR_LEVEL_CRITICAL,
			      0,
			      ERROR_ADVISE_ABORT,
			      "Caught unknown exception.", "");
    }
}
HBCI_Bank *HBCI_API_bankFactory(const HBCI_API *h,
				int country,
				const char *bankCode,
				const char *server)
{
    assert(h);
    HBCI::Pointer<HBCI_Bank> bp = h->bankFactory(country, 
						 string(bankCode ? bankCode : ""), 
						 string(server ? server : ""));
    bp.setAutoDelete(false);
    return bp.ptr();
}
HBCI_Medium* HBCI_API_mediumFactory(const HBCI_API *h,
				    const char *name,
				    MediumType mediumType)
{
    assert(h);
    try {
	HBCI::Pointer<HBCI_Medium> bp = 
	    h->mediumFactory(string((name ? name : "")), 
			     mediumType);
	bp.setAutoDelete(false);
	return bp.ptr();
    }
    catch(HBCI_Error xerr) {
	fprintf(stderr,"HBCI_API_mediumFactory: Caught error: %s\n",
		xerr.errorString().c_str());
	return 0;
    }
    catch (...) {
	fprintf(stderr, "HBCI_API_mediumFactory: Caught unknown exception.\n");
	return 0;
    }
}
HBCI_User *HBCI_API_userFactory(HBCI_Bank *b,
				HBCI_Medium *m,
				int mediumAutoDelete,
				const char *userid)
{
    assert(b);
    assert(m);
    HBCI::Pointer<HBCI_Medium> mp = m;
    mp.setAutoDelete(mediumAutoDelete);
    HBCI::Pointer<HBCI_User> up = 
	HBCI_API::userFactory(bankPointer(b), 
			      mp, 
			      string(userid ? userid : ""));
    up.setAutoDelete(false);
    return up.ptr();
}
HBCI_Customer *HBCI_API_customerFactory(HBCI_User *u,
					const char *id,
					const char *custName)
{
    assert(u);
    HBCI::Pointer<HBCI_Customer> cp = 
	HBCI_API::customerFactory(userPointer(u),
				  string(id ? id : ""),
				  string(custName ? custName : ""));
    cp.setAutoDelete(false);
    return cp.ptr();
}
HBCI_Account *HBCI_API_accountFactory(HBCI_Bank *b,
				      const char *accountId,
				      const char *accountSubId)
{
    assert(b);
    HBCI::Pointer<HBCI_Account> ap = 
	(accountSubId ? 
	 HBCI_API::accountFactory(bankPointer(b),
				  string(accountId ? accountId : ""),
				  string(accountSubId ? accountSubId : ""))
	 : (accountId ?
	    HBCI_API::accountFactory(bankPointer(b),
				     string(accountId ? accountId : ""))
	    : HBCI_API::accountFactory(bankPointer(b))));

    ap.setAutoDelete(false);
    return ap.ptr();
}
HBCI_Medium *HBCI_API_createNewMedium(const HBCI_API *h,
				      int country,
				      const char *bankId,
				      const char *userid,
				      const char *name,
                                      MediumType mt,
				      HBCI_Error **err)
{
    assert(h);
    assert(err);
    try {
	HBCI::Pointer<HBCI::Medium> mp =
	    h->createNewMedium(country, 
			       string(bankId ? bankId : ""), 
			       string(userid ? userid : ""),
			       string(name ? name : ""), 
			       mt);
	mp.setAutoDelete(false);
	return mp.ptr();
    }
    catch(HBCI_Error xerr) {
	fprintf(stderr,"HBCI_API_createNewMedium: Caught error: %s\n",
		xerr.errorString().c_str());
	*err = new HBCI_Error(xerr);
	return 0;
    }
    catch (...) {
	fprintf(stderr, "HBCI_API_createNewMedium: Caught unknown exception.\n");
	*err = new HBCI_Error("HBCI_API_createNewMedium",
			      ERROR_LEVEL_CRITICAL,
			      0,
			      ERROR_ADVISE_ABORT,
			      "Caught unknown exception.", "");
	return 0;
    }
}

HBCI_Error *HBCI_API_loadEnvironment(HBCI_API *h,
				     const char *filename)
{
    assert(h);
    try {
	return new HBCI_Error(h->loadEnvironment(string(filename ? filename : "")));
    }
    catch (HBCI_Error err) {
	fprintf(stderr,"HBCI_API_loadEnvironment: Caught error: %s\n",
		err.errorString().c_str());
	return new HBCI_Error(err);
    }
    catch (...) {
	fprintf(stderr, "HBCI_API_loadEnvironment: Caught unknown exception.\n");
	return new HBCI_Error("HBCI_API_loadEnvironment",
			      ERROR_LEVEL_CRITICAL,
			      0,
			      ERROR_ADVISE_ABORT,
			      "Caught unknown exception.", "");
    }
}
HBCI_Error *HBCI_API_saveEnvironment(const HBCI_API *h,
				     const char *filename)
{
    assert(h);
    try {
	return new HBCI_Error(h->saveEnvironment(string(filename ? filename : "")));
    }
    catch (HBCI_Error err) {
	fprintf(stderr,"HBCI_API_saveEnvironment: Caught error: %s\n",
		err.errorString().c_str());
	return new HBCI_Error(err);
    }
    catch (...) {
	fprintf(stderr, "HBCI_API_saveEnvironment: Caught unknown exception.\n");
	return new HBCI_Error("HBCI_API_saveEnvironment",
			      ERROR_LEVEL_CRITICAL,
			      0,
			      ERROR_ADVISE_ABORT,
			      "Caught unknown exception.", "");
    }
}
HBCI_Error *HBCI_API_addJob(HBCI_API *h, HBCI_OutboxJob *j)
{
    assert(h);
    try {
	return new HBCI_Error(h->addJob(j));
    }
    catch (HBCI_Error err) {
	fprintf(stderr,"HBCI_API_addJob: Caught error: %s\n",
		err.errorString().c_str());
	return new HBCI_Error(err);
    }
    catch (...) {
	fprintf(stderr, "HBCI_API_addJob: Caught unknown exception.\n");
	return new HBCI_Error("HBCI_API_addJob",
			      ERROR_LEVEL_CRITICAL,
			      0,
			      ERROR_ADVISE_ABORT,
			      "Caught unknown exception.", "");
    }
}
HBCI_Error *HBCI_API_executeQueue(HBCI_API *h, int changesAllowed)
{
    assert(h);
    try {
	return new HBCI_Error(h->executeQueue(changesAllowed));
    }
    catch (HBCI_Error err) {
	fprintf(stderr,"HBCI_API_executeQueue: Caught error: %s\n",
		err.errorString().c_str());
	return new HBCI_Error(err);
    }
    catch (...) {
	fprintf(stderr, "HBCI_API_executeQueue: Caught unknown exception.\n");
	return new HBCI_Error("HBCI_API_executeQueue",
			      ERROR_LEVEL_CRITICAL,
			      0,
			      ERROR_ADVISE_ABORT,
			      "Caught unknown exception.", "");
    }
}
void HBCI_API_clearQueueByResult(HBCI_API *h, OutboxJob_Result result)
{
    assert(h);
    h->clearQueueByResult(result);
}
void HBCI_API_clearQueueByStatus(HBCI_API *h, OutboxJob_Status status)
{
    assert(h);
    h->clearQueueByStatus(status);
}
int HBCI_API_totalAccounts(const HBCI_API *h)
{
    assert(h);
    return h->totalAccounts();
}
int HBCI_API_totalUsers(const HBCI_API *h)
{
    assert(h);
    return h->totalUsers();
}
