/***************************************************************************
 $RCSfile: mediumrdh.h,v $
                             -------------------
    cvs         : $Id: mediumrdh.h,v 1.6 2002/12/27 23:06:55 aquamaniac Exp $
    begin       : Sun Nov 18 2001
    copyright   : (C) 2001 by Martin Preuss
    email       : openhbci@aquamaniac.de

 ***************************************************************************
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

/** @file mediumrdh.h
 *
 * @short Key file medium HBCI::MediumRDH and its C wrapper @ref
 * HBCI_MediumRDH.  */


#ifndef HBCIMEDIUMRDH_H
#define HBCIMEDIUMRDH_H


#ifdef __cplusplus
#include <string>
#endif /* __cplusplus */
#include <openhbci/medium.h>
#include <openhbci/mediumrdhbase.h>
#ifdef __cplusplus
#include <openhbci/cryptkey.h>
#include <openhbci/rsakey.h>


namespace HBCI {

/**
 * @short Implements access to a RDH medium (key file) in OpenHBCI.
 *
 * This class saves it's data in a file using the TLV format.
 * That format is:
 * - (T)ag    (first byte, describing the type of data)
 * - (L)ength (two bytes, tells the length of the data following)
 * - (V)alue  (contains the data itself)
 * To find the position of the next tag just add the (L)ength value plus 3
 * (the (T)ag and (L)ength field) to the current position.
 *
 * Please note that some tag numbers are treated specially:
 * - numbers >0xc0 are assumed to have subtags
 * - numbers 0x00 and 0xc0 are assumed to be descriptive tags and therefore
 *   ignored by this library (that number is used by external programs which
 *   show the content of a medium file. The (V)alue of descriptive tags
 *   (0x00) has in its first byte the type of data:
 *   - "&" stands for binary data
 *   - "$" stands for ASCII text
 *   Well, descriptive tags are not used by this class neither created,
 *   but maybe I decide (or someone else does) to write a little program to
 *   show the content of the media files so it is better to think forward ;-)
 *
 * (FIXME: Need some comment here on the compatibility with other HBCI
 * applications. Do other HBCI applications use the same file format
 * (such that the MediumRDH can easily use their key files) or do all
 * other HBCI applications use a proprietary format?)
 *
 * @author Martin Preuss<openhbci@aquamaniac.de> */
class DLLIMPORT MediumRDH: public MediumRDHBase {
    friend class RDHConverter;
 private:
    /**
     * @short Internal MediumRDH data
     * @internal
     */
    struct userData {
        Pointer<RSAKey> userPubSignatureKey;
        Pointer<RSAKey> userPrivateSignatureKey;
        Pointer<RSAKey> userPubCryptKey;
        Pointer<RSAKey> userPrivateCryptKey;
        string userId;
        userData();
        ~userData();
    };

    /**
     * @short Internal MediumRDH data
     * @internal
     */
    struct instData {
        int country;
        string instcode;
        string systemid;
        string addr;
        Pointer<RSAKey> instPubSignKey;
        Pointer<RSAKey> instPubCryptKey;
        list<string> users;
        instData();
        ~instData();
    };

    // admin stuff
    string _path;
    string _pin;

    list<userData*> lusers;
    list<instData*> linst;
    userData *_currentUser;
    instData *_currentInst;

    int _mountcount;
    unsigned int _seq;
    Pointer<User> _mounter;

    Pointer<RSAKey> _readKey(const string &data);
    userData *_readUser(const string &data);
    instData *_readInst(const string &data);
    string _writeKey(Pointer<RSAKey> key) const;
    string _writeUser(userData &ud) const;
    string _writeInst(instData &id) const;
    userData *_findUser(const string &userid) const;
    bool _hasUser(const instData *id, const string &userid) const;
    instData *_findInst(const string &instid);

    bool _getContext(int num,
                     instData &id,
                     userData &ud,
                     unsigned int &seq) const;

    /**
     * Reads the medii information from a file.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    void readFile(const string &path, const string &pin);

    /**
     * Writes the medii information into a file.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    void writeFile(const string &path, const string &pin);

    /**
     * Returns the institutes public crypt key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    Pointer<RSAKey> instPubCryptKey() const;

    /**
     * Returns the institutes public signature key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    Pointer<RSAKey> instPubSignKey() const;

 public:
    /**
     * Constructor.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @param hbci pointer to the HBCI object your program uses.
     * @param path full path to the file which serves as a medium. All keys are stored in this
     * file.
     */
    MediumRDH(const Hbci *hbci, const string &path);

    ~MediumRDH();

    /** @name Medium management */
    //@{
    /**
     * Since some media (like hbci chip cards) are able to store multiple
     * account entries you have to select one.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    bool selectContext(int country,
                       const string &instcode,
                       const string &userid);

    /**
     * Creates a context in your medium. A context is just the combination of a userid and the
     * institute code (+ country code).
     * IMPORTANT: If you want to add a context to an already existing
     * medium you MUST call this method with the medium MOUNTED !!
     * Otherwise the keys on that medium will be LOST !!
     * Well, of course, if that medium does not exist you have to this method
     * without prior mounting (since there is nothing to mount).
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    bool createContext(int country,
                       const string &instcode,
                       const string &userid,
                       const string &pin="");

    bool createMedium(int country,
                      const string &instcode,
                      const string &userid,
                      const string &pin="") { return false;};

    /**
     * Mounts a medium and makes its crypto methods available.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @return true if medium mounted
     * @param user customer that wants to mount this medium
     * @param pin The secret PIN that is needed to actually mount this 
     * medium (FIXME: is this correct?)
     */
    Error mountMedium(Pointer<User> user, const string &pin="");

    /**
     * Unmounts a medium so that it may be removed.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @return true if medium unmounted
     */
    void unmountMedium(const string &pin="");

    unsigned int nextSEQ();
    void resetSEQ();

    //@}

    /** @name Medium Information */
    //@{
    /**
     * Checks whether the medium is mounted.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @return true if medium unmounted, false if not (or on error)
     */
    bool isMounted();

    /**
     * Returns the id of the medium. For DDV this is the CID of the chip
     * card, for RDH this is the system ID assigned to us by an institute.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    string mediumId() const;

    /**
     * Returns the security mode of this medium (DDV, RDH)
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    int securityMode() const { return HBCI_SECURITY_RDH;};

    /**
     * Returns the id of the medium.
     * For DDV this is the card number, for RDH this is the name of the file.
     */
    const string& mediumName() const { return _path; };

    /**
     * Returns the type this medium is of.
     *
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    MediumType mediumType() const { return MediumRDHFileOld;};

    /**
     * Returns the number of the sign key. RDH-media should return
     * the number of the users private sign key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    int signKeyNumber() const;

    /**
     * Returns the version of the crypt key. RDH-media should return
     * the version of the users private sign key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    int signKeyVersion() const;

    /**
     * Returns the number of the crypt key. RDH-media should return
     * the number of the institutes public crypt key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    int cryptKeyNumber() const;

    /**
     * Returns the version of the crypt key. RDH-media should return
     * the version of the institutes public crypt key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    int cryptKeyVersion() const;

    /**
     * Returns the owner of the crypt key. RDH-media should return
     * the owner of the institutes public crypt key. This is retrieved from
     * HIISA segments.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    const string &cryptKeyOwner() const;

    /**
     * Returns the context with the given number.
     * Some media are able of storing multiple contexts (like chip cards).
     * Please note that the medium has to be mounted prior to calling this
     * method.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @return true if a context with that number exists, false on error
     * @param num number of the context (starting with 1)
     * @param countrycode reference to an int var to receive the country code
     * of the context (280 for Germany)
     * @param instcode reference to a string variable to receive the institute
     * code of that context (German "Bankleitzahl")
     * @param userid reference to a string variable to receive the user id of
     * the context (assigned to you by the institute)
     * @param server reference to a string to receive the server address
     */
    bool getContext(int num,
                    int &countrycode,
                    string &instcode,
                    string &userid,
                    string &server) const;
    //@}
    
    /** @name Cryptographic Methods */
    //@{
    /**
     * Creates a key for encryption of data.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @return false on error
     */
    string createMessageKey() const;

    /**
     * Lets the card encrypt the key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @return false on error
     * @param srckey the key to encode
     */
    string encryptKey(const string &srckey);

    /**
     * Lets the card decrypt the key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @return false on error
     * @param srckey the key to decode
     */
    string decryptKey(const string &srckey);

    /**
     * Verify the signature of given data
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @return false on error
     * @param data data whose signatur is to be verified
     * @param signature signature to compare against
     */
    bool verify(const string &data, const string &signature);

    /**
     * sign data
     * @author Martin Preuss<openhbci@aquamaniac.de>
     * @return false on error
     * @param data data whose signatur is to be created
     */
    string sign(const string &data);

    string encrypt(const string &src, const string &key);

    string decrypt(const string &src, const string &key);
    //@}

    /** @name Key Generation and management */
    //@{
    /**
     * Generates the users sign and crypt keys.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    virtual bool createUserKeys(bool activate=true);

    /**
     * Activates the previously created keys. This is <b>required</b> if
     * you specified "false" as arguement to @ref createUserKeys().
     * This will finally store the created keys to their destination, so that
     * they are usable for further signing/crypting.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    virtual bool activateKeys() { return false;};

    /**
     * This returns the temporary sign key which has been created by
     * @ref createUserKeys(). You'll need this in a key change protocol.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    virtual Pointer<RSAKey> getTempSignKey() { return 0;};

    /**
     * This returns the temporary crypt key which has been created by
     * @ref createUserKeys(). You'll need this in a key change protocol.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    virtual Pointer<RSAKey> getTempCryptKey() { return 0;};


    /*
     * Sets the institutes public keys.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    //bool setInstituteKeys(Pointer<RSAKey> signkey,
    //                    Pointer<RSAKey> cryptkey);

    /**
     * Sets the institutes public crypt key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    bool setInstituteCryptKey(Pointer<RSAKey> cryptkey);

    /**
     * Sets the institutes public sign key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    bool setInstituteSignKey(Pointer<RSAKey> signkey);

    /**
     * Sets the SystemId.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    void setSystemId(const string &newid);
    //@}

    /** @name Key Information */
    //@{
    /** Returns true, if the institute/bank of this MediumRDH uses a
     * public signature key. */
    bool hasInstSignKey() const;

    /**
     * Returns the users public crypt key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    Pointer<RSAKey> userPubCryptKey() const;

    /**
     * Returns the users public sign key.
     * @author Martin Preuss<openhbci@aquamaniac.de>
     */
    Pointer<RSAKey> userPubSignKey() const;
    //@}

    /** @name Ini-Letter Information */
    //@{
    /** Returns the modulus of the institute key that is needed to
     * print the ini-letter.
     *
     * @param crypt If true, then the data of <i>crypt</i> key is read
     * instead of that of the <i>sign</i> key. This exists for the
     * case when your institute does not sign its messages. In that
     * not-so-good case, MediumRDH::hasInstSignKey returned false (the
     * inverse of this argument). If unsure, leave it to the default
     * value <code>false</code>.  */
    string getInstIniLetterModulus(bool crypt=false) const;

    /**
     * return the ini-letter-exponent that is needed to print the ini-letter
     *
     * @param crypt If true, then the data of <i>crypt</i> key is read
     * instead of that of the <i>sign</i> key. This exists for the
     * case when your institute does not sign its messages. In that
     * not-so-good case, MediumRDH::hasInstSignKey returned false (the
     * inverse of this argument). If unsure, leave it to the default
     * value <code>false</code>.  */
    string getInstIniLetterExponent(bool crypt=false) const;

    /**
     * return the ini-letter-hash that is needed to print the ini-letter
     *
     * @param crypt If true, then the data of <i>crypt</i> key is read
     * instead of that of the <i>sign</i> key. This exists for the
     * case when your institute does not sign its messages. In that
     * not-so-good case, MediumRDH::hasInstSignKey returned false (the
     * inverse of this argument). If unsure, leave it to the default
     * value <code>false</code>. 
     */
    string getInstIniLetterHash(bool crypt=false) const;

    /**
     * Returns the number of the institute key.
     *
     * @param usecrypt If true, then the data of <i>crypt</i> key is read
     * instead of that of the <i>sign</i> key. This exists for the
     * case when your institute does not sign its messages. In that
     * not-so-good case, MediumRDH::hasInstSignKey returned false (the
     * inverse of this argument). If unsure, leave it to the default
     * value <code>false</code>. 
     */
    int getInstKeyNumber(bool usecrypt = false) const;
    /**
     * Returns the version of the institute key.
     *
     * @param usecrypt If true, then the data of <i>crypt</i> key is read
     * instead of that of the <i>sign</i> key. This exists for the
     * case when your institute does not sign its messages. In that
     * not-so-good case, MediumRDH::hasInstSignKey returned false (the
     * inverse of this argument). If unsure, leave it to the default
     * value <code>false</code>. 
     */
    int getInstKeyVersion(bool usecrypt = false) const;

    /**
     * return the ini-letter-modulus that is needed to print the ini-letter
     */
    string getUserIniLetterModulus() const;

    /**
     * return the ini-letter-exponent that is needed to print the ini-letter
     */
    string getUserIniLetterExponent() const;

    /**
     * return the ini-letter-hash that is needed to print the ini-letter
     */
    string getUserIniLetterHash() const;

    /**
     * This method is needed when generating the users ini letter.
     * @return the number of the user's public crypt key
     */
    int getUserKeyNumber() const;

    /**
     * This method is needed when generating the users ini letter.
     * @return the version of the user's public crypt key
     */
    int getUserKeyVersion() const;
    //@}

};

} /* namespace HBCI */

typedef HBCI::MediumRDH HBCI_MediumRDH;
extern "C" {
#else /* __cplusplus */
    typedef struct HBCI_MediumRDH HBCI_MediumRDH;
#endif /* __cplusplus */
    /** Downcast. Returns NULL if this is not a MediumRDH. */
    extern HBCI_MediumRDH *HBCI_Medium_MediumRDH(HBCI_Medium *h);
    extern int HBCI_MediumRDH_hasInstSignKey(const HBCI_MediumRDH *h);
    extern char *
    HBCI_MediumRDH_getInstIniLetterExponent(const HBCI_MediumRDH *h, 
					    int usecrypt);
    extern char *
    HBCI_MediumRDH_getInstIniLetterHash(const HBCI_MediumRDH *h, 
					int usecrypt);
    extern char *
    HBCI_MediumRDH_getInstIniLetterModulus(const HBCI_MediumRDH *h, 
					   int usecrypt);
    extern int HBCI_MediumRDH_getInstKeyNumber(const HBCI_MediumRDH *h, 
					       int usecrypt);
    extern int HBCI_MediumRDH_getInstKeyVersion(const HBCI_MediumRDH *h, 
						int usecrypt);
    extern char *
    HBCI_MediumRDH_getUserIniLetterExponent(const HBCI_MediumRDH *h);
    extern char *
    HBCI_MediumRDH_getUserIniLetterHash(const HBCI_MediumRDH *h);
    extern char *
    HBCI_MediumRDH_getUserIniLetterModulus(const HBCI_MediumRDH *h);
    extern int HBCI_MediumRDH_getUserKeyNumber(const HBCI_MediumRDH *h);
    extern int HBCI_MediumRDH_getUserKeyVersion(const HBCI_MediumRDH *h);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif

